import numpy as np

__all__ = ["Spectrum"]


class Spectrum(np.ndarray):
    """
    Class representing a 1 dimensional spectrum.

    Attributes
    ----------
    freq_axis : `~numpy.ndarray`
        One-dimensional array with the frequency values.
    data : `~numpy.ndarray`
        One-dimensional array which the intensity at a particular frequency at every data-point.

    Examples
    --------
    >>> from radiospectra.spectrum import Spectrum
    >>> import numpy as np
    >>> data = np.linspace(1, 100, 100)
    >>> freq_axis = np.linspace(0, 10, 100)
    >>> spec = Spectrum(data, freq_axis)
    >>> spec.peek()   # doctest: +SKIP
    """

    def __new__(cls, data, *args, **kwargs):
        return np.asarray(data).view(cls)

    def __init__(self, data, freq_axis):
        if np.shape(data)[0] != np.shape(freq_axis)[0]:
            raise ValueError("Dimensions of data and frequency axis do not match")
        self.freq_axis = freq_axis

    def plot(self, axes=None, **matplot_args):
        """
        Plot spectrum onto current axes.

        Parameters
        ----------
        axes : `~matplotlib.axes.Axes` or `None`
            If provided the spectrum will be plotted on the given axes.
            Else the current `matplotlib` axes will be used.
        **matplot_args : dict
            Any additional plot arguments that should be used
            when plotting.

        Returns
        -------
         `~matplotlib.axes.Axes`
            The plot axes.
        """
        from matplotlib import pyplot as plt

        # Get current axes
        if not axes:
            axes = plt.gca()
        params = {}
        params.update(matplot_args)
        lines = axes.plot(self.freq_axis, self, **params)
        return lines

    def peek(self, **matplot_args):
        """
        Plot spectrum onto a new figure.
        Parameters
        ----------
        **matplot_args : dict
            Any additional plot arguments that should be used when plotting.

        Returns
        -------
        `~matplotlib.Figure`
            A plot figure.

        Examples
        --------
        >>> from radiospectra.spectrum import Spectrum
        >>> import numpy as np
        >>> spec = Spectrum(np.linspace(1, 100, 100), np.linspace(0, 10, 100))
        >>> spec.peek()  # doctest: +SKIP
        """
        from matplotlib import pyplot as plt

        figure = plt.figure()
        self.plot(**matplot_args)
        figure.show()
        return figure



# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# flake8: NOQA: E402

# -- stdlib imports ------------------------------------------------------------
import os
import datetime
from packaging.version import Version

# -- Read the Docs Specific Configuration --------------------------------------

# This needs to be done before radiospectra is imported
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if on_rtd:
    os.environ["SUNPY_CONFIGDIR"] = "/home/docs/"
    os.environ["HOME"] = "/home/docs/"
    os.environ["LANG"] = "C"
    os.environ["LC_ALL"] = "C"
    os.environ["HIDE_PARFIVE_PROGESS"] = "True"


# -- Project information -----------------------------------------------------
# The full version, including alpha/beta/rc tags
from radiospectra import __version__  # NOQA

_version = Version(__version__)
version = release = str(_version)
# Avoid "post" appearing in version string in rendered docs
if _version.is_postrelease:
    version = release = _version.base_version
# Avoid long githashes in rendered Sphinx docs
elif _version.is_devrelease:
    version = release = f"{_version.base_version}.dev{_version.dev}"
is_development = _version.is_devrelease
is_release = not(_version.is_prerelease or _version.is_devrelease)

project = "radiospectra"
author = "The SunPy Community"
copyright = f"{datetime.datetime.now().year}, {author}"  # noqa: A001

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "matplotlib.sphinxext.plot_directive",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
    "sphinx_changelog",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
html_extra_path = ["robots.txt"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# Treat everything in single ` as a Python reference.
default_role = "py:obj"

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"python": ("https://docs.python.org/", None)}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sunpy"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]

# By default, when rendering docstrings for classes, sphinx.ext.autodoc will
# make docs with the class-level docstring and the class-method docstrings,
# but not the __init__ docstring, which often contains the parameters to
# class constructors across the scientific Python ecosystem. The option below
# will append the __init__ docstring to the class-level docstring when rendering
# the docs. For more options, see:
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autoclass_content
autoclass_content = "both"

# -- Other options ----------------------------------------------------------



class PcolormeshPlotMixin:
    """
    Class provides plotting functions using `~pcolormesh`.
    """

    def plot(self, axes=None, **kwargs):
        """
        Plot the spectrogram.

        Parameters
        ----------
        axes : `matplotlib.axis.Axes`, optional
            The axes where the plot will be added.
        kwargs :
            Arguments pass to the plot call `pcolormesh`.

        Returns
        -------
        `matplotlib.collections.QuadMesh`
        """
        import matplotlib.dates as mdates
        from matplotlib import pyplot as plt

        if axes is None:
            fig, axes = plt.subplots()
        else:
            fig = axes.get_figure()

        if hasattr(self.data, "value"):
            data = self.data.value
        else:
            data = self.data

        title = f"{self.observatory}, {self.instrument}"
        if self.instrument != self.detector:
            title = f"{title}, {self.detector}"

        axes.set_title(title)
        axes.plot(self.times.datetime[[0, -1]], self.frequencies[[0, -1]], linestyle="None", marker="None")
        if self.times.shape[0] == self.data.shape[0] and self.frequencies.shape[0] == self.data.shape[1]:
            ret = axes.pcolormesh(self.times.datetime, self.frequencies.value, data, shading="auto", **kwargs)
        else:
            ret = axes.pcolormesh(self.times.datetime, self.frequencies.value, data[:-1, :-1], shading="auto", **kwargs)
        axes.set_xlim(self.times.datetime[0], self.times.datetime[-1])
        locator = mdates.AutoDateLocator(minticks=4, maxticks=8)
        formatter = mdates.ConciseDateFormatter(locator)
        axes.xaxis.set_major_locator(locator)
        axes.xaxis.set_major_formatter(formatter)
        fig.autofmt_xdate()
        # Set current axes/image if pyplot is being used (makes colorbar work)
        for i in plt.get_fignums():
            if axes in plt.figure(i).axes:
                plt.sca(axes)
                plt.sci(ret)
        return ret


class NonUniformImagePlotMixin:
    """
    Class provides plotting functions using `NonUniformImage`.
    """

    def plotim(self, fig=None, axes=None, **kwargs):
        import matplotlib.dates as mdates
        from matplotlib import pyplot as plt
        from matplotlib.image import NonUniformImage

        if axes is None:
            fig, axes = plt.subplots()

        im = NonUniformImage(axes, interpolation="none", **kwargs)
        im.set_data(mdates.date2num(self.times.datetime), self.frequencies.value, self.data)
        axes.images.append(im)



from pathlib import Path
from unittest import mock

import numpy as np
import pytest

from sunpy.net import attrs as a
from sunpy.net.fido_factory import Fido

from radiospectra.net.attrs import PolType
from radiospectra.net.sources.eovsa import EOVSAClient

MOCK_PATH = "sunpy.net.scraper.urlopen"


@pytest.fixture
def client():
    return EOVSAClient()


@pytest.fixture
def http_responses():
    paths = [Path(__file__).parent / "data" / n for n in ["eovsa_resp1.html", "eovsa_resp2.html"]]
    response_htmls = []
    for p in paths:
        with p.open("r") as f:
            response_htmls.append(f.read())

    return response_htmls


@mock.patch(MOCK_PATH)
def test_client(urlopen, client, http_responses):
    urlopen.return_value.read = mock.MagicMock()
    urlopen.return_value.read.side_effect = http_responses
    urlopen.close = mock.MagicMock(return_value=None)
    query = client.search(a.Time("2020/10/05 00:00", "2020/10/06 23:00"), a.Instrument("EOVSA"))
    assert urlopen.call_count == 2
    # last call should be for 2020/10/06
    assert urlopen.call_args[0][0] == "https://ovsa.njit.edu/fits/synoptic/2020/10/06/"
    assert len(query) == 4


@mock.patch(MOCK_PATH)
def test_client_observatory(urlopen, client, http_responses):
    urlopen.return_value.read = mock.MagicMock()
    urlopen.return_value.read.side_effect = http_responses
    urlopen.close = mock.MagicMock(return_value=None)
    query = client.search(a.Time("2020/10/05 00:00", "2020/10/06 00:00"), a.Instrument("EOVSA"), PolType.cross)
    assert urlopen.call_count == 2
    # last call should be for 2020/10/06
    assert urlopen.call_args[0][0] == "https://ovsa.njit.edu/fits/synoptic/2020/10/06/"
    assert len(query) == 2
    assert np.all(query["PolType"] == "Cross")


@pytest.mark.remote_data
def test_fido():
    query = Fido.search(a.Time("2020/10/05 00:00", "2020/10/06 00:00"), a.Instrument("EOVSA"), PolType.cross)
    assert len(query[0]) == 2
    assert np.all(query[0]["PolType"] == "Cross")



from sunpy.net.attr import SimpleAttr

__all__ = ["Spacecraft", "Observatory", "PolType"]


class Spacecraft(SimpleAttr):
    """
    The STEREO Spacecraft A (Ahead) or B (Behind).
    """


class Observatory(SimpleAttr):
    """
    Observatory.
    """


class PolType(SimpleAttr):
    """
    Polarisation Type.
    """



"""
``radiospectra`` Spectrum
"""

from radiospectra.spectrogram.sources import *  # NOQA
from radiospectra.spectrogram.spectrogram_factory import *  # NOQA
from radiospectra.spectrogram.spectrogrambase import *  # NOQA



from sunpy.net import attrs as a
from sunpy.net.dataretriever.client import GenericClient, QueryResponse
from sunpy.net.scraper import Scraper
from sunpy.time.timerange import TimeRange

from radiospectra.net.attrs import Observatory

__all__ = ["RSTNClient"]


class RSTNClient(GenericClient):
    """
    Radio Spectrometer Telescope Network (RSTN) hosted at NOAA
    `National Geophysical Data <https://www.ngdc.noaa.gov>`__ (NGDC) archive.

    Examples
    --------
    >>> from radiospectra import net
    >>> from sunpy.net import Fido, attrs as a
    >>> query = Fido.search(a.Time('2003/03/15 00:00', '2003/03/15 23:59'),
    ...                     a.Instrument('RSTN'), net.Observatory('San Vito'))  #doctest: +REMOTE_DATA
    >>> query  #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...
    Results from 1 Provider:
    <BLANKLINE>
    1 Results from the RSTNClient:
           Start Time               End Time        Provider Instrument Observatory
    ----------------------- ----------------------- -------- ---------- -----------
    2003-03-15 00:00:00.000 2003-03-15 23:59:59.999     RSTN       RSTN    San Vito
    <BLANKLINE>
    <BLANKLINE>
    """

    baseurl = (
        r"https://www.ngdc.noaa.gov/stp/space-weather/solar-data/"
        r"solar-features/solar-radio/rstn-spectral/{obs}/%Y/%m/.*.gz"
    )
    pattern = r"{}/rstn-spectral/{obs}/{year:4d}/{month:2d}/" r"{obs_short:2l}{year2:2d}{month2:2d}{day:2d}.SRS.gz"

    observatory_map = {
        "Holloman": "holloman",
        "Learmonth": "learmonth",
        "Palehua": "palehua",
        "Sagamore Hill": "sagamore",
        "San Vito": "san-vito",
    }
    observatory_map = {**observatory_map, **dict(map(reversed, observatory_map.items()))}

    def search(self, *args, **kwargs):
        baseurl, pattern, matchdict = self.pre_search_hook(*args, **kwargs)
        metalist = []
        for obs in matchdict["Observatory"]:
            scraper = Scraper(baseurl.format(obs=self.observatory_map[obs.title()]), regex=True)
            tr = TimeRange(matchdict["Start Time"], matchdict["End Time"])
            filesmeta = scraper._extract_files_meta(tr, extractor=pattern, matcher=matchdict)

            for i in filesmeta:
                rowdict = self.post_search_hook(i, matchdict)
                metalist.append(rowdict)

        return QueryResponse(metalist, client=self)

    def post_search_hook(self, exdict, matchdict):
        original = super().post_search_hook(exdict, matchdict)
        obs, *_ = (original.pop(name) for name in ["obs", "year2", "month2", "obs_short"])
        original["Observatory"] = self.observatory_map[obs]
        return original

    @classmethod
    def register_values(cls):
        adict = {
            a.Provider: [("RSTN", "Radio Solar Telescope Network.")],
            a.Instrument: [("RSTN", "Radio Solar Telescope Network.")],
            Observatory: [
                ("Holloman", "Holloman"),
                ("Learmonth", "Learmonth"),
                ("Palehua", "Palehua"),
                ("Sagamore Hill", "Sagamore Hill"),
                ("San Vito", "San Vito"),
            ],
        }
        return adict



from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram

__all__ = [
    "ILOFARMode357Spectrogram",
]


class ILOFARMode357Spectrogram(GenericSpectrogram):
    """
    Irish LOFAR Station mode 357 Spectrogram
    """

    @property
    def mode(self):
        return self.meta.get("mode")

    @property
    def polarisation(self):
        return self.meta.get("polarisation")

    @classmethod
    def is_datasource_for(cls, data, meta, **kwargs):
        return meta["instrument"] == "ILOFAR"



from sunpy.net import attrs as a
from sunpy.net.dataretriever.client import GenericClient

from radiospectra.net.attrs import PolType


class EOVSAClient(GenericClient):
    """
    Provides access to `Extended Owens Valley Solar Array <http://www.ovsa.njit.edu>`__ (EOVSA) data.

    Examples
    --------
    >>> from radiospectra import net
    >>> from sunpy.net import Fido, attrs as a
    >>> query = Fido.search(a.Time('2020/10/05 00:00', '2020/10/06 00:00'),
    ...                     a.Instrument('EOVSA'), net.PolType.cross)  #doctest: +REMOTE_DATA
    >>> query  #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    2 Results from the EOVSAClient:
           Start Time               End Time        Provider Instrument PolType
    ----------------------- ----------------------- -------- ---------- -------
    2020-10-05 00:00:00.000 2020-10-05 23:59:59.999    EOVSA      EOVSA   Cross
    2020-10-06 00:00:00.000 2020-10-06 23:59:59.999    EOVSA      EOVSA   Cross
    <BLANKLINE>
    <BLANKLINE>
    """
    from sunpy import __version__
    if __version__ >= "6.1.0":
        pattern = ("https://ovsa.njit.edu/fits/synoptic/{{year:4d}}/{{month:2d}}/{{day:2d}}/"
                "EOVSA_{{PolType:5l}}_{{year:4d}}{{month:2d}}{{day:2d}}.fts")
    else:
        baseurl = "https://ovsa.njit.edu/fits/synoptic/%Y/%m/%d/EOVSA_.*_%Y%m%d.fts"
        pattern = "{}/synoptic/{year:4d}/{month:2d}/{day:2d}/EOVSA_{PolType:5l}_{year:4d}{month:2d}{day:2d}.fts"
    pol_map = {"Total": "TPall", "Cross": "XPall", "TPall": "Total", "XPall": "Cross"}

    @classmethod
    def pre_search_hook(cls, *args, **kwargs):
        baseurl, pattern, matchdict = super().pre_search_hook(*args, **kwargs)
        pol_values = [cls.pol_map[p.capitalize()] for p in matchdict["PolType"]]
        matchdict["PolType"] = pol_values
        return baseurl, pattern, matchdict

    def post_search_hook(self, exdict, matchdict):
        original = super().post_search_hook(exdict, matchdict)
        original["PolType"] = self.pol_map[original["PolType"]]
        return original

    @classmethod
    def register_values(cls):
        adict = {
            a.Provider: [("EOVSA", "EOVSA")],
            a.Instrument: [("EOVSA", "ExtendedOwens Valley Solar Array.")],
            PolType: [("Total", "Total polarisation"), ("Cross", "Cross polarisation")],
        }
        return adict



import astropy.units as u
from astropy.coordinates.earth import EarthLocation

from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram

__all__ = ["CALISTOSpectrogram"]


class CALISTOSpectrogram(GenericSpectrogram):
    """
    CALISTO Spectrogram from the e-CALISTO network.

    Examples
    --------
    >>> import radiospectra.net
    >>> from sunpy.net import Fido, attrs as a
    >>> from radiospectra.spectrogram import Spectrogram
    >>> from radiospectra.net import attrs as ra
    >>> query = Fido.search(a.Time('2019/10/05 23:00', '2019/10/06 00:59'),  #doctest: +REMOTE_DATA
    ...                     a.Instrument('eCALLISTO'), ra.Observatory('ALASKA'))  #doctest: +REMOTE_DATA
    >>> downloaded = Fido.fetch(query[0][0])  #doctest: +REMOTE_DATA
    >>> spec = Spectrogram(downloaded[0])  #doctest: +REMOTE_DATA
    >>> spec  #doctest: +REMOTE_DATA
    <CALISTOSpectrogram ALASKA, E-CALLISTO, E-CALLISTO 215000.0 kHz - 418937.98828125 kHz, 2019-10-05T23:00:00.757 to 2019-10-05T23:15:00.000>
    >>> spec.plot()  #doctest: +REMOTE_DATA
    <matplotlib.collections.QuadMesh object at ...>
    """

    def __init__(self, data, meta, **kwargs):
        super().__init__(meta=meta, data=data, **kwargs)

    @property
    def observatory_location(self):
        lat = self.meta["fits_meta"]["OBS_LAT"] * u.deg * 1.0 if self.meta["fits_meta"]["OBS_LAC"] == "N" else -1.0
        lon = self.meta["fits_meta"]["OBS_LON"] * u.deg * 1.0 if self.meta["fits_meta"]["OBS_LOC"] == "E" else -1.0
        height = self.meta["fits_meta"]["OBS_ALT"] * u.m
        return EarthLocation(lat=lat, lon=lon, height=height)

    @classmethod
    def is_datasource_for(cls, data, meta, **kwargs):
        return meta["instrument"] == "e-CALLISTO" or meta["detector"] == "e-CALLISTO"



import astropy.units as u

from sunpy.net import attrs as a
from sunpy.net.dataretriever.client import GenericClient, QueryResponse
from sunpy.net.scraper import Scraper
from sunpy.time.timerange import TimeRange

__all__ = ["RFSClient"]

RECEIVER_FREQUENCIES = {
    "rfs_lfr": a.Wavelength(10 * u.kHz, 1.7 * u.MHz),
    "rfs_hfr": a.Wavelength(1.3 * u.MHz, 19.2 * u.MHz),
}


class RFSClient(GenericClient):
    """
    Provides access to Parker Solar Probe FIELDS Radio Frequency Spectrometer data
    `archive <https://spdf.gsfc.nasa.gov/pub/data/psp/fields/>`__ at
    `NASA Goddard Space Physics Data Facility (SPDF) <https://spdf.gsfc.nasa.gov>`__.

    Examples
    --------
    >>> import radiospectra.net
    >>> from sunpy.net import Fido, attrs as a
    >>> results = Fido.search(a.Time("2019/10/02", "2019/10/05"),
    ...                        a.Instrument('rfs'))  #doctest: +REMOTE_DATA
    >>> results  #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    8 Results from the RFSClient:
    <BLANKLINE>
           Start Time               End Time        Instrument Source Provider     Wavelength
                                                                                      kHz
    ----------------------- ----------------------- ---------- ------ -------- -----------------
    2019-10-02 00:00:00.000 2019-10-02 23:59:59.999        RFS    PSP     SPDF    10.0 .. 1700.0
    2019-10-03 00:00:00.000 2019-10-03 23:59:59.999        RFS    PSP     SPDF    10.0 .. 1700.0
    2019-10-04 00:00:00.000 2019-10-04 23:59:59.999        RFS    PSP     SPDF    10.0 .. 1700.0
    2019-10-05 00:00:00.000 2019-10-05 23:59:59.999        RFS    PSP     SPDF    10.0 .. 1700.0
    2019-10-02 00:00:00.000 2019-10-02 23:59:59.999        RFS    PSP     SPDF 1300.0 .. 19200.0
    2019-10-03 00:00:00.000 2019-10-03 23:59:59.999        RFS    PSP     SPDF 1300.0 .. 19200.0
    2019-10-04 00:00:00.000 2019-10-04 23:59:59.999        RFS    PSP     SPDF 1300.0 .. 19200.0
    2019-10-05 00:00:00.000 2019-10-05 23:59:59.999        RFS    PSP     SPDF 1300.0 .. 19200.0
    <BLANKLINE>
    <BLANKLINE>
    """

    baseurl = (
        r"https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/{Wavelength}/"
        r"{year}/psp_fld_l2_(\w){{7}}_(\d){{8}}_v(\d){{2}}.cdf"
    )
    pattern = r"{}/{Wavelength}/{year:4d}/" r"psp_fld_l2_{Wavelength}_{year:4d}{month:2d}{day:2d}_v{:2d}.cdf"

    @classmethod
    def _check_wavelengths(cls, wavelength):
        """
        Check for overlap between given wavelength and receiver frequency coverage
        defined in ``RECEIVER_FREQUENCIES``.

        Parameters
        ----------
        wavelength : `sunpy.net.attrs.Wavelength`
            Input wavelength range to check

        Returns
        -------
        `list`
            List of receivers names or empty list if no overlap
        """
        # Input wavelength range is completely contained in one receiver range
        receivers = [k for k, v in RECEIVER_FREQUENCIES.items() if wavelength in v]
        # If not defined need to continue
        if not receivers:
            # Overlaps but not contained in, either max in lfr or min hfr
            if wavelength.min in RECEIVER_FREQUENCIES["rfs_hfr"] or wavelength.max in RECEIVER_FREQUENCIES["rfs_hfr"]:
                receivers.append("rfs_hfr")
            if wavelength.min in RECEIVER_FREQUENCIES["rfs_lfr"] or wavelength.max in RECEIVER_FREQUENCIES["rfs_lfr"]:
                receivers.append("rfs_lfr")
            # min in lfr and max in hfr
            # min and max of combined lft and hfr contained in give wavelength range
            if a.Wavelength(RECEIVER_FREQUENCIES["rfs_lfr"].min, RECEIVER_FREQUENCIES["rfs_hfr"].max) in wavelength:
                receivers = ["rfs_lfr", "rfs_hfr"]
            # If we get here the is no overlap so set to empty list
        return receivers

    def search(self, *args, **kwargs):
        """
        Query this client for a list of results.

        Parameters
        ----------
        *args: `tuple`
            `sunpy.net.attrs` objects representing the query.
        **kwargs: `dict`
             Any extra keywords to refine the search.

        Returns
        -------
        A `QueryResponse` instance containing the query result.
        """
        matchdict = self._get_match_dict(*args, **kwargs)
        req_wave = matchdict.get("Wavelength", None)
        receivers = RECEIVER_FREQUENCIES.keys()
        if req_wave is not None:
            receivers = self._check_wavelengths(req_wave)

        metalist = []
        start_year = matchdict["Start Time"].datetime.year
        end_year = matchdict["End Time"].datetime.year
        tr = TimeRange(matchdict["Start Time"], matchdict["End Time"])
        for receiver in receivers:
            for year in range(start_year, end_year + 1):
                urlpattern = self.baseurl.format(Wavelength=receiver, year=year)
                scraper = Scraper(urlpattern, regex=True)
                filesmeta = scraper._extract_files_meta(tr, extractor=self.pattern)
                for i in filesmeta:
                    rowdict = self.post_search_hook(i, matchdict)
                    metalist.append(rowdict)

        return QueryResponse(metalist, client=self)

    def post_search_hook(self, exdict, matchdict):
        """
        This method converts 'rfs_hfr' and 'rfs_lfr' in the url's metadata to the
        frequency ranges for low and high frequency receivers.
        """
        rowdict = super().post_search_hook(exdict, matchdict)
        if rowdict["Wavelength"] == "rfs_hfr":
            fr = RECEIVER_FREQUENCIES["rfs_hfr"]
            rowdict["Wavelength"] = u.Quantity([float(fr.min.value), float(fr.max.value)], unit=fr.unit)
        elif rowdict["Wavelength"] == "rfs_lfr":
            fr = RECEIVER_FREQUENCIES["rfs_lfr"]
            rowdict["Wavelength"] = u.Quantity([float(fr.min.value), float(fr.max.value)], unit=fr.unit)
        return rowdict

    @classmethod
    def register_values(cls):
        adict = {
            a.Instrument: [("RFS", ("Radio Frequency Spectrometer"))],
            a.Source: [("PSP", "Parker Solar Probe")],
            a.Provider: [("SPDF", "NASA Goddard Space Physics Data Facility")],
            a.Wavelength: [("*")],
        }
        return adict



from pathlib import Path
from unittest import mock

import pytest

import astropy.units as u

from sunpy.net import Fido
from sunpy.net import attrs as a

from radiospectra.net.attrs import PolType
from radiospectra.net.sources.ilofar import ILOFARMode357Client


@pytest.fixture
def client():
    return ILOFARMode357Client()


@pytest.fixture
def html_responses():
    paths = [Path(__file__).parent / "data" / n for n in ["ilofar_resp1.html", "ilofar_resp2.html"]]
    response_htmls = []
    for p in paths:
        with p.open("r") as f:
            response_htmls.append(f.read())

    return response_htmls


@mock.patch("sunpy.net.scraper.urlopen")
def test_ilofar_client(mock_urlopen, client, html_responses):
    mock_urlopen.return_value.read = mock.MagicMock()
    mock_urlopen.return_value.read.side_effect = html_responses * 2
    mock_urlopen.close = mock.MagicMock(return_value=None)
    atr = a.Time("2018/06/01", "2018/06/02")
    query = client.search(atr)

    called_urls = [
        "https://data.lofar.ie/2018/06/01/bst/kbt/rcu357_1beam/",
        "https://data.lofar.ie/2018/06/02/bst/kbt/rcu357_1beam/",
        "https://data.lofar.ie/2018/06/01/bst/kbt/rcu357_1beam_datastream/",
        "https://data.lofar.ie/2018/06/02/bst/kbt/rcu357_1beam_datastream/",
    ]
    assert called_urls == [call[0][0] for call in mock_urlopen.call_args_list]
    assert len(query) == 8
    assert query[0]["Source"] == "ILOFAR"
    assert query[0]["Provider"] == "ILOFAR"
    assert query[0]["Start Time"].iso == "2018-06-01 10:00:41.000"
    assert query[0]["Polarisation"] == "X"


@mock.patch("sunpy.net.scraper.urlopen")
def test_ilofar_client_polarisation(mock_urlopen, client, html_responses):
    mock_urlopen.return_value.read = mock.MagicMock()
    mock_urlopen.return_value.read.side_effect = html_responses * 2
    mock_urlopen.close = mock.MagicMock(return_value=None)
    atr = a.Time("2018/06/01", "2018/06/02")
    query_x = client.search(atr, PolType("X"))
    assert len(query_x) == 4
    assert query_x[0]["Source"] == "ILOFAR"
    assert query_x[0]["Provider"] == "ILOFAR"
    assert query_x[0]["Start Time"].iso == "2018-06-01 10:00:41.000"
    assert query_x[0]["Polarisation"] =='X'


@mock.patch("sunpy.net.scraper.urlopen")
def test_ilofar_client_wavelength(mock_urlopen, client, html_responses):
    mock_urlopen.return_value.read = mock.MagicMock()
    mock_urlopen.return_value.read.side_effect = html_responses * 6
    mock_urlopen.close = mock.MagicMock(return_value=None)
    atr = a.Time("2018/06/01", "2018/06/02")
    query_both_low = client.search(atr, a.Wavelength(1 * u.MHz, 5 * u.MHz))
    query_both_high = client.search(atr, a.Wavelength(1 * u.GHz, 2 * u.GHz))

    assert len(query_both_low) == 0
    assert len(query_both_high) == 0

    query_low_in = client.search(atr, a.Wavelength(90 * u.MHz, 1 * u.GHz))
    query_both_in = client.search(atr, a.Wavelength(15 * u.MHz, 230 * u.MHz))
    query_high_in = client.search(atr, a.Wavelength(5 * u.MHz, 90 * u.MHz))

    for query in [query_low_in, query_both_in, query_high_in]:
        assert len(query) == 8


@pytest.mark.remote_data
def test_fido():
    atr = a.Time("2018/06/01", "2018/06/02")
    query = Fido.search(atr, a.Instrument("ILOFAR"))

    assert isinstance(query[0].client, ILOFARMode357Client)
    query = query[0]
    assert len(query) == 8
    assert query[0]["Source"] == "ILOFAR"
    assert query[0]["Provider"] == "ILOFAR"
    assert query[0]["Start Time"].iso == "2018-06-01 10:00:41.000"
    assert query[0]["Polarisation"] == "X"


@pytest.mark.remote_data
def test_fido_other_dataset():
    atr = a.Time("2021/08/01", "2021/10/01")
    query = Fido.search(atr, a.Instrument("ILOFAR"))

    assert isinstance(query[0].client, ILOFARMode357Client)
    query = query[0]
    assert len(query) == 38



from pathlib import Path
from datetime import datetime
from unittest import mock

import numpy as np

import astropy.units as u
from astropy.time import Time

from sunpy.net import attrs as a

from radiospectra.spectrogram import Spectrogram
from radiospectra.spectrogram.sources import SWAVESSpectrogram


@mock.patch("radiospectra.spectrogram.spectrogram_factory.parse_path")
def test_swaves_lfr(parse_path_moc):
    meta = {
        "instrument": "swaves",
        "observatory": "STEREO A",
        "product": "average",
        "start_time": Time("2020-11-28 00:00:00"),
        "end_time": Time("2020-11-28 23:59:00"),
        "wavelength": a.Wavelength(2.6 * u.kHz, 153.4 * u.kHz),
        "detector": "lfr",
        "freqs": [
            2.6,
            2.8,
            3.1,
            3.4,
            3.7,
            4.0,
            4.4,
            4.8,
            5.2,
            5.7,
            6.2,
            6.8,
            7.4,
            8.1,
            8.8,
            9.6,
            10.4,
            11.4,
            12.4,
            13.6,
            14.8,
            16.1,
            17.6,
            19.2,
            20.9,
            22.8,
            24.9,
            27.1,
            29.6,
            32.2,
            35.2,
            38.3,
            41.8,
            45.6,
            49.7,
            54.2,
            59.1,
            64.5,
            70.3,
            76.7,
            83.6,
            91.2,
            99.4,
            108.4,
            118.3,
            129.0,
            140.6,
            153.4,
        ]
        * u.kHz,
        "times": np.arange(1440) * u.min,
    }
    array = np.zeros((48, 1440))
    parse_path_moc.return_value = [(array, meta)]
    file = Path("fake.dat")
    spec = Spectrogram(file)
    assert isinstance(spec, SWAVESSpectrogram)
    assert spec.observatory == "STEREO A"
    assert spec.instrument == "SWAVES"
    assert spec.detector == "LFR"
    assert spec.start_time.datetime == datetime(2020, 11, 28, 0, 0)
    assert spec.end_time.datetime == datetime(2020, 11, 28, 23, 59)
    assert spec.wavelength.min == 2.6 * u.kHz
    assert spec.wavelength.max == 153.4 * u.kHz


@mock.patch("radiospectra.spectrogram.spectrogram_factory.parse_path")
def test_swaves_hfr(parse_path_moc):
    meta = {
        "instrument": "swaves",
        "observatory": "STEREO A",
        "product": "average",
        "start_time": Time("2020-11-28 00:00:00"),
        "end_time": Time("2020-11-28 23:59:00"),
        "wavelength": a.Wavelength(125.0 * u.kHz, 16025.0 * u.kHz),
        "detector": "hfr",
        "freqs": np.linspace(125, 16025, 319) * u.kHz,
        "times": np.arange(1440) * u.min,
    }
    array = np.zeros((319, 1440))
    parse_path_moc.return_value = [(array, meta)]
    file = Path("fake.dat")
    spec = Spectrogram(file)
    assert isinstance(spec, SWAVESSpectrogram)
    assert spec.observatory == "STEREO A"
    assert spec.instrument == "SWAVES"
    assert spec.detector == "HFR"
    assert spec.start_time.datetime == datetime(2020, 11, 28, 0, 0)
    assert spec.end_time.datetime == datetime(2020, 11, 28, 23, 59)
    assert spec.wavelength.min == 125 * u.kHz
    assert spec.wavelength.max == 16025 * u.kHz



from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram

__all__ = ["RSTNSpectrogram"]


class RSTNSpectrogram(GenericSpectrogram):
    """
    Radio Solar Telescope Network.

    Examples
    --------
    >>> import radiospectra.net
    >>> from sunpy.net import Fido, attrs as a
    >>> from radiospectra.spectrogram import Spectrogram
    >>> query = Fido.search(a.Time('2017/09/07 00:00', '2017/09/07 23:00'), a.Instrument.rstn)  #doctest: +REMOTE_DATA
    >>> downloaded = Fido.fetch(query[0][0])  #doctest: +REMOTE_DATA
    >>> spec = Spectrogram(downloaded[0])  #doctest: +REMOTE_DATA
    >>> spec  #doctest: +REMOTE_DATA
    <RSTNSpectrogram LEARMONTH, RSTN, RSTN 25000.0 kHz - 180000.0 kHz, 2017-09-06T22:31:51.000 to 2017-09-07T10:06:36.000>
    >>> spec.plot()  #doctest: +REMOTE_DATA
    <matplotlib.collections.QuadMesh object at ...>
    """

    @classmethod
    def is_datasource_for(cls, data, meta, **kwargs):
        return meta["instrument"] == "RSTN"
